@@ -11,12 +11,29 @@ package SQL::Statement;
 use strict;
 use SQL::Parser;
 use SQL::Eval;
-use vars qw($VERSION $numexp $s2pops $arg_num $dlm);
+use vars qw($VERSION $numexp $s2pops $arg_num $dlm $warg_num $HAS_DBI);
+BEGIN {
+    eval { require 'DBI.pm' };
+    $HAS_DBI = 1 unless $@;
+    *is_number = ($HAS_DBI)
+               ? *DBI::looks_like_number
+	       : sub {
+                     my @strs = @_;
+                     for my $x(@strs) {
+                         return 0 if !defined $x;
+                         return 0 if $x !~ $numexp;
+                     }
+                     return 1;
+                 };
+}
 
-$VERSION = '1.004';
+#use locale;
+
+$VERSION = '1.06';
 
 $dlm = '~';
 $arg_num=0;
+$warg_num=0;
 $s2pops = {
               'LIKE'   => {'s'=>'LIKE',n=>'LIKE'},
               'CLIKE'  => {'s'=>'CLIKE',n=>'CLIKE'},
@@ -33,6 +50,7 @@ BEGIN {
     sub qr {}
   }
 }
+
 sub new {
     my $class  = shift;
     my $sql    = shift;
@@ -56,9 +74,9 @@ sub new {
     my $parser_dialect = $flags->{"dialect"} || 'AnyData';
     $parser_dialect = 'AnyData' if $parser_dialect =~ /^(CSV|Excel)$/;
 
-#    if (!ref($parser) or (ref($parser) and ref($parser) !~ /^SQL::Parser/)) {
-    if (!ref($parser)) {
-        # print "NEW PARSER\n";
+    if (!ref($parser) or (ref($parser) and ref($parser) !~ /^SQL::Parser/)) {
+ #   if (!ref($parser)) {
+#         print "NEW PARSER\n";
         $parser = new SQL::Parser($parser_dialect,$flags);
     }
 #       unless ref $parser and ref $parser =~ /^SQL::Parser/;
@@ -111,7 +129,7 @@ sub prepare {
        #print $self->command, " [$max_placeholders]\n";
        if ($max_placeholders) {
            for my $i(0..$max_placeholders-1) {
-               SQL::Statement::Param->new($i);
+               $self->{"params"}->[$i] = SQL::Statement::Param->new($i);
            }
        }
        if ($self->{"sort_spec_list"}) {
@@ -159,27 +177,48 @@ sub execute {
     $self->{'params'}= $params;
     my($table, $msg);
     my($command) = $self->command();
-    return undef unless $command;
+    return $self->do_err( 'No command found!') unless $command;
     ($self->{'NUM_OF_ROWS'}, $self->{'NUM_OF_FIELDS'},
-     $self->{'data'}) = $self->$command($data, $params);
-    my $tables;
+          $self->{'data'}) = $self->$command($data, $params);
+#=pod
     my $names = $self->{NAME};
-    @{$self->{NAME}} = map {
-      my $org = $self->{ORG_NAME}->{$_};
-      $org =~ s/^"//;
-      $org =~ s/"$//;
-      $org =~ s/""/"/g;
-      $org;
-    } @$names;
+    @$names = map {
+        my $org = $self->{ORG_NAME}->{$_}; # from the file header
+        $org =~ s/^"//;
+        $org =~ s/"$//;
+        $org =~ s/""/"/g;
+        $org;
+    } @$names  if $self->{asterisked_columns};
+    $names = $self->{org_col_names} unless $self->{asterisked_columns};
+
+    $self->{NAME} = $names;
+if ($command eq 'SELECT') {
+#use mylibs; zwarn $self;
+#print "\n\n";
+#print "[" . ($self->{asterisked_columns}||'') . "]";
+#print "~@$names~" if $names;
+#print "\n\n";
+    }
+#=cut
+    my $tables;
     @$tables = map {$_->{"name"}} @{ $self->{"tables"} };
     delete $self->{'tables'};  # Force closing the tables
-      for (@$tables) {
-           push @{ $self->{"tables"} },
-                SQL::Statement::Table->new($_);
-       }
+    for (@$tables) {
+        push @{ $self->{"tables"} }, SQL::Statement::Table->new($_);
+    }
     $self->{'NUM_OF_ROWS'} || '0E0';
 }
 
+sub CONNECT ($$$) {
+    my($self, $data, $params) = @_;
+    if ($self->can('open_connection')) {
+        my $dsn = $self->{connection}->{dsn};
+        my $tbl = $self->{connection}->{tbl};
+        return $self->open_connection($dsn,$tbl,$data,$params)
+    }
+    (0, 0);
+}
+
 sub CREATE ($$$) {
     my($self, $data, $params) = @_;
     my($eval,$foo) = $self->open_tables($data, 1, 1);
@@ -197,6 +236,12 @@ sub CREATE ($$$) {
 
 sub DROP ($$$) {
     my($self, $data, $params) = @_;
+    if ($self->{ignore_missing_table}) {
+         eval { $self->open_tables($data,0,0) };
+         if ($@ and $@ =~ /no such (table|file)/i ) {
+             return (-1,0);
+	 }
+    }
     my($eval) = $self->open_tables($data, 0, 1);
 #    return undef unless $eval;
     return (-1,0) unless $eval;
@@ -217,7 +262,6 @@ sub INSERT ($$$) {
     $table->seek($data, 0, 2);
     my($array) = [];
     my($val, $col, $i);
-###    my($columns) = $table->{'colNums'};
     my($cNum) = scalar($self->columns());
     my $param_num = 0;
     if ($cNum) {
@@ -227,40 +271,17 @@ sub INSERT ($$$) {
             $val = $self->row_values($i);
             if ($val and ref($val) eq 'SQL::Statement::Param') {
                 $val = $eval->param($val->num());
-            }
-            elsif ($val and $val->{type} eq 'placeholder') {
+          }
+          elsif ($val and $val->{type} eq 'placeholder') {
                 $val = $eval->param($param_num++);
 	    }
             else {
 	         $val = $self->get_row_value($val,$eval);
 	    }
-#print $col->name."~".$self->{ORG_NAME}->{$col->name}."\n";
-#	    my $ccnum = $table->column_num($col->name);
-#	    if (!defined $ccnum) {
-#                $ccnum = $table->column_num(qq/"$col->name"/);
-#	    }
-#            $array->[$ccnum] = $val;
             $array->[$table->column_num($col->name())] = $val;
-#            $array->[$table->column_num(lc $col->name())] = $val;
-
-#if ( $self->{ORG_NAME}->{$col->name} ) {
-#    $col = qq/"/.$self->{ORG_NAME}->{$col}.qq/"/;
-#} 
-#            $array->[$table->column_num($col->name())] = $val;
         }
     } else {
         return $self->do_err("Bad col names in INSERT");
-        # INSERT INTO $table VALUES (value, ...)
-        # NOT USED BECAUSE $cNum automatically assigned
-#        $cNum = scalar($self->row_values());
-#        for ($i = 0;  $i < $cNum;  $i++) {
-#            $val = $self->row_values($i);
-#            $val = $self->get_row_value($val);
-#            if (ref($val) eq 'SQL::Statement::Param') {
-#                $val = $eval->param($val->num());
-#            }
-#            $array->[$i] = $val;
-#        }
     }
     $table->push_row($data, $array);
     (1, 0);
@@ -275,6 +296,17 @@ sub DELETE ($$$) {
     my($table) = $eval->table($self->tables(0)->name());
     my($affected) = 0;
     my(@rows, $array);
+    if ( $table->can('delete_one_row') ) {
+        while (my $array = $table->fetch_row($data)) {
+            if ($self->eval_where($eval,'',$array)) {
+                ++$affected;
+                $array = $self->{fetched_value} if $self->{fetched_from_key};
+                $table->delete_one_row($data,$array);
+                return ($affected, 0) if $self->{fetched_from_key};
+	      }
+        }
+        return ($affected, 0);
+    }
     while ($array = $table->fetch_row($data)) {
         if ($self->eval_where($eval,'',$array)) {
             ++$affected;
@@ -310,9 +342,12 @@ sub UPDATE ($$$) {
     my($table) = $eval->table($self->tables(0)->name());
     my $tname = $self->tables(0)->name();
     my($affected) = 0;
-    my(@rows, $array, $val, $col, $i);
+    my(@rows, $array, $f_array, $val, $col, $i);
     while ($array = $table->fetch_row($data)) {
         if ($self->eval_where($eval,$tname,$array)) {
+            if( $self->{fetched_from_key} and $table->can('update_one_row') ){
+                $array = $self->{fetched_value};
+            }
         my $param_num =$arg_num;
         #print $param_num;
         #print $eval->param($param_num); print "@$params"; exit;
@@ -341,6 +376,10 @@ sub UPDATE ($$$) {
             }
             ++$affected;
         }
+        if ($self->{fetched_from_key}){
+            $table->update_one_row($data,$array);
+            return ($affected, 0);
+        }
         push(@rows, $array);
     }
     $table->seek($data, 0, 0);
@@ -461,14 +500,15 @@ sub JOIN {
     $tableB = shift @tables;
     my $tableAobj = $eval->table($tableA);
     my $tableBobj = $eval->table($tableB);
+#use mylibs; print $tableA; zwarn $tableAobj; exit;
     $tableAobj->{"REAL_NAME"} = $tableAobj->{"NAME"} ||= $tableA;
     $tableBobj->{"REAL_NAME"} = $tableBobj->{"NAME"} ||= $tableB;
     if (my $aliasA = $self->{table_alias}->{$tableA}) {
         $tableAobj->{"NAME"} = shift @{$self->{table_alias}->{$tableA}};
-    } 
+    }
     if (my $aliasB = $self->{table_alias}->{$tableB}) {
-        $tableBobj->{"NAME"} = shift @{$self->{table_alias}->{$tableB}};
-    } 
+        $tableBobj->{"NAME"} =  shift @{$self->{table_alias}->{$tableB}} ;
+    }
 
 #use mylibs; zwarn $self;
 #die "$tableA,$tableB";
@@ -692,6 +732,7 @@ sub SELECT ($$) {
         }
         if ($col eq '*') {
             $ar = $table->col_names();
+
 #@$ar = map {lc $_} @$ar;
             for ($i = 0;  $i < @$ar;  $i++) {
                 my $cName = $ar->[$i];
@@ -760,7 +801,7 @@ sub SELECT ($$) {
 		  }
 	    }
             $tbl ||= $self->colname2table($col);
-            #print "$tbl~$col\n";
+            #print "$tbl~\n";
             next if exists($columns{$tbl}->{"$col"});
             $pos = $table->column_num($col) unless defined $pos;
             push(@extraSortCols, $pos);
@@ -773,6 +814,7 @@ sub SELECT ($$) {
     }
     while (my $array = $table->fetch_row($data)) {
         if ($self->eval_where($e,$tableName,$array)) {
+            $array = $self->{fetched_value} if $self->{fetched_from_key};
             # Note we also include the columns from @extraSortCols that
             # have to be ripped off later!
             my @row;
@@ -782,10 +824,12 @@ sub SELECT ($$) {
             #            else {
                 @extraSortCols = () unless @extraSortCols;
             #print "[$_]" for @$cList; print "\n";
-###new
+
             @row = map { defined $_ and defined $array->[$_] ? $array->[$_] : undef } (@$cList, @extraSortCols);
-            #	    }
             push(@$rows, \@row);
+            return (scalar(@$rows),scalar @{$self->{column_names}},$rows)
+ 	        if $self->{fetched_from_key};
+            #	    }
         }
     }
     if (@order_by) {
@@ -814,8 +858,8 @@ sub SELECT ($$) {
                     $result = defined $d ? -1 : 0;
                 } elsif (!defined($d)) {
                     $result = 1;
-#	        } elsif ( is_number($c) && is_number($d) ) {
-	        } elsif ( $c =~ $numexp && $d =~ $numexp ) {
+	        } elsif ( is_number($c,$d) ) {
+#	        } elsif ( $c =~ $numexp && $d =~ $numexp ) {
                     $result = ($c <=> $d);
                 } else {
   		    if ($self->{"case_fold"}) {
@@ -899,7 +943,7 @@ sub SELECT ($$) {
             my $sf_index = -1;
  	    for my $sf(@{$self->{"set_function"}}) {
               $sf_index++;
-	      if ($sf->{"arg"} eq '*') {
+	      if ($sf->{arg} and $sf->{"arg"} eq '*') {
                   $final_row[$sf_index]++;
 	      }
               else {
@@ -908,19 +952,29 @@ sub SELECT ($$) {
                 next unless defined $v;
                 my $final = $final_row[$sf_index];
                 $final++      if $name =~ /COUNT/;
-                $final += $v  if $name =~ /SUM|AVG/;
+#                $final += $v  if $name =~ /SUM|AVG/;
+                if( $name =~ /SUM|AVG/) {
+                    return $self->do_err("Can't use $name on a string!")
+#                      unless $v =~ $numexp;
+                      unless is_number($v);
+                    $final += $v;
+                }
                 #
                 # Thanks Dean Kopesky dean.kopesky@reuters.com
                 # submitted patch to make MIN/MAX do cmp on strings
                 # and == on numbers
                 #
-                $final  = $v  if !defined $final
+                # but thanks also to Michael Kovacs mkovacs@turing.une.edu.au
+                # for catching a problem when a MAX column is 0
+                # necessitating !$final instead of ! defined $final
+                #
+                $final  = $v  if !$final
                               or ( $name eq 'MAX'
                                    and $v
                                    and $final
                                    and anycmp($v,$final) > 0
                                  );
-                $final  = $v  if !defined $final
+                $final  = $v  if !$final
                               or ( $name eq 'MIN' 
                                    and defined $v
                                    and anycmp($v,$final) < 0
@@ -943,12 +997,12 @@ sub anycmp($$) {
     my ($a,$b) = @_;
     $a = '' unless defined $a;
     $b = '' unless defined $b;
-    return ($a =~ $numexp && $b =~ $numexp)
+#    return ($a =~ $numexp && $b =~ $numexp)
+    return ( is_number($a,$b) )
         ? ($a <=> $b)
         : ($a cmp $b);
 }
 
-
 sub eval_where {
     my $self   = shift;
     my $eval   = shift;
@@ -1017,7 +1071,8 @@ sub process_predicate {
 	    }
             else {
                 if (defined $val1 and defined $val2 and $op !~ /^IS/i ) {
-                    $op = ( $val1 =~ $numexp && $val2 =~ $numexp )
+#                    $op = ( $val1 =~ $numexp && $val2 =~ $numexp )
+                    $op = ( is_number($val1,$val2) )
                         ? $s2pops->{"$op"}->{'n'}
                         : $s2pops->{"$op"}->{'s'};
                 }
@@ -1025,12 +1080,26 @@ sub process_predicate {
 	}
         else {
             if (defined $val1 and defined $val2 and $op !~ /^IS/i ) {
-                $op = ( $val1 =~ $numexp && $val2 =~ $numexp )
+#                $op = ( $val1 =~ $numexp && $val2 =~ $numexp )
+                $op = ( is_number($val1,$val2) )
                     ? $s2pops->{"$op"}->{'n'}
                     : $s2pops->{"$op"}->{'s'};
 	    }
         }
 #        print "[$val1] [$op] [$val2]\n";
+        my $neg = $pred->{"neg"};
+        my($table) = $eval->table($self->tables(0)->name());
+    if ( $pred->{op} eq '=' and !$neg and $table->can('fetch_one_row')
+       ) {
+        my $key_col = $table->fetch_one_row(1,1);
+        if ($pred->{arg1}->{value} =~ /^$key_col$/i) {
+            $self->{fetched_from_key}=1;
+            $self->{fetched_value} = $table->fetch_one_row(
+                0,$val2
+            );
+            return 1;
+	}
+    }
         my $match = $self->is_matched($val1,$op,$val2) || 0;
         if ($pred->{"neg"}) {
            $match = $match ? 0 : 1;
@@ -1039,12 +1108,6 @@ sub process_predicate {
     }
 }
 
-sub is_number {
-    my $x=shift;
-    return 0 if !defined $x;
-    return 1 if $x =~ /^([+-]?)(?=\d|\.\d)\d*(\.\d*)?([Ee]([+-]?\d+))?$/;
-    return 0;
-}
 sub is_matched {
     my($self,$val1,$op,$val2)=@_;
     #print "[$val1] [$op] [$val2]\n";
@@ -1065,7 +1128,7 @@ sub is_matched {
     if ($op =~ /LIKE|CLIKE/i) {
         $val2 = quotemeta($val2);
         $val2 =~ s/\\%/.*/g;
-        $val2 =~ s/_/?/g;
+        $val2 =~ s/_/./g;
     }
     if ( !$self->{"alpha_compare"} && $op =~ /lt|gt|le|ge/ ) {
         return 0;
@@ -1092,38 +1155,34 @@ sub open_tables {
     my($self, $data, $createMode, $lockMode) = @_;
     my @call = caller 4;
     my $caller = $call[3];
-    $caller =~ s/^([^:]*::[^:]*)::.*$/$1/;
+    if ($caller) {
+        $caller =~ s/^([^:]*::[^:]*)::.*$/$1/;
+    }
     my @c;
     my $t;
     my $is_col;
     my @tables = $self->tables;
+    my $count=-1;
     for ( @tables) {
+        $count++;
         my $name = $_->{"name"};
         undef $@;
-         eval{
-    ## statement change
-    #           $t->{"$name"} =
-    #              $self->open_table($data, $name, $createMode, $lockMode);
-        my $open_name = $name;
-#        $open_name =~ s/^"(.+)"$/$1/;
-#$name = lc $name;
+        eval{
+            my $open_name = $self->{org_table_names}->[$count];
            if ($caller && $caller =~ /^DBD::AnyData/) {
                $caller .= '::Statement' if $caller !~ /::Statement/;
-               $t->{"$name"} = $caller->open_table($data, $open_name, $createMode,
-                                              $lockMode);
+               $t->{"$name"} = $caller->open_table($data, $open_name,
+                                                   $createMode, $lockMode);
 	   }
            else {
-               $t->{"$name"} = $self->open_table($data, $open_name, $createMode, 
-                                               $lockMode);
+               $t->{"$name"} = $self->open_table($data, $open_name,
+                                                 $createMode, $lockMode);
 	   }
 
-    ####
 	};
         my $err = $t->{"$name"}->{errstr};
         return $self->do_err($err) if $err;
         return $self->do_err($@) if $@;
-        return(SQL::Eval->new({'tables' => $t}), [])
-            if $self->command eq 'DROP';
 my @cnames;
 for my $c(@{$t->{"$name"}->{"col_names"}}) {
   my $newc;
@@ -1137,15 +1196,13 @@ for my $c(@{$t->{"$name"}->{"col_names"}}) {
   }
    push @cnames, $newc;
    $self->{ORG_NAME}->{$newc}=$c;
-#   push @{$self->{ORG_NAME}},$c;
-
 }
 my $col_nums;
 my $i=0;
 for (@cnames) {
   $col_nums->{$_} = $i++;
 }
-$t->{"$name"}->{"col_nums"}  = $col_nums;
+$t->{"$name"}->{"col_nums"}  = $col_nums; # upper cased
 $t->{"$name"}->{"col_names"} = \@cnames;
 #use mylibs; zwarn $t->{$name};
         my $tcols = $t->{"$name"}->col_names;
@@ -1160,7 +1217,9 @@ $t->{"$name"}->{"col_names"} = \@cnames;
 	}
         @c = ( @c, @newcols );
     }
-#use mylibs; zwarn $t->{test}->{org_cols};
+    my $all_cols = $self->{all_cols} || [];
+    @$all_cols = (@$all_cols,@c);
+    $self->{all_cols} = $all_cols;
     return SQL::Eval->new({'tables' => $t}), \@c;
 }
 
@@ -1181,7 +1240,8 @@ sub verify_columns {
 ###z
     for my $c(@tmpcols) {
        if ($c->{"name"} eq '*' and defined $c->{"table"}) {
-
+          return $self->do_err("Can't find table ". $c->{"table"})
+              unless $eval->{"tables"}->{$c->{"table"}};
           my $tcols = $eval->{"tables"}->{$c->{"table"}}->col_names;
 # @$tcols = map{lc $_} @$tcols ;
           return $self->do_err("Couldn't find column names!")
@@ -1202,8 +1262,10 @@ sub verify_columns {
     @tmpcols = map {$_->{name}} @$usr_cols;
 #     @tmpcols = map {lc $_->{name}} @$usr_cols;
     my $fully_qualified_cols=[];
+
     my %col_exists   = map {$_=>1} @tmp_cols;
-#    my %col_exists   = map {lc $_=>1} @tmp_cols;
+
+
     my %short_exists = map {s/^([^.]*)\.(.*)/$1/; $2=>$1} @tmp_cols;
     my(%is_member,@duplicates,%is_duplicate);
     @duplicates = map {s/[^.]*\.(.*)/$1/; $_} @$all_cols;
@@ -1230,6 +1292,7 @@ sub verify_columns {
           $table = $table->name;
        }
 ###endnew
+#print "Content-type: text/html\n\n"; print $self->command; print "$col!!!<p>";
        if ( $col eq '*' and $num_tables == 1) {
           $table ||= $self->tables->[0]->{"name"};
           if (ref $table eq 'SQL::Statement::Table') {
@@ -1290,7 +1353,7 @@ sub verify_columns {
                return $self->do_err("Ambiguous column name '$c'")
                    if $is_duplicate{$c};
                return $self->do_err("No such column '$c'")
-               unless $short_exists{"$c"};
+                      unless $short_exists{"$c"} or ($c !~ /^"/ and $short_exists{"\U$c"});
                $table = $short_exists{"$c"};
                $col   = $c;
            }
@@ -1303,10 +1366,12 @@ sub verify_columns {
 #print qq/$table."$col"/;
 # use mylibs; zwarn $self->{ORG_NAME};
 	     }
-             
+#use mylibs; zwarn \%col_exists;
+#print "<$table . $col>";
                return $self->do_err("No such column '$table.$col'")
 #                     unless $col_exists{"\L$table.$col"};
                      unless $col_exists{"$table.$col"}
+                      or $col_exists{"\L$table.".$col};
 ;#                         or $col_exists{qq/$table."/.$self->{ORG_NAME}->{$col}.qq/"/}
 ;
            }
@@ -1326,10 +1391,12 @@ sub verify_columns {
               $self->{"columns"} = \@newcols;
        }
     }
-#use mylibs; zwarn $self->{columns};
+#use mylibs; zwarn $fully_qualified_cols;
+
     return $fully_qualified_cols;
 }
 
+
 sub distinct {
     my $self = shift;
     return 1 if $self->{"set_quantifier"}
@@ -1337,7 +1404,6 @@ sub distinct {
     return 0;
 }
 
-
 sub command { shift->{"command"} }
 
 sub params {
@@ -1364,7 +1430,7 @@ sub row_values {
         return $self->{"values"}->[$val_num];
     }
     if (wantarray) {
-        return map{$_->{"values"} } @{$self->{"values"}};
+        return map{$_->{"value"} } @{$self->{"values"}};
     }
     else {
         return scalar @{ $self->{"values"} };
@@ -1425,7 +1491,8 @@ sub get_row_value {
                my $val = $self->get_row_value($vals[$i],$eval,$rowhash);
                return $self->do_err(
                    qq{Bad numeric expression '$vals[$i]->{"value"}'!}
-               ) unless defined $val and $val =~ $numexp;
+#               ) unless defined $val and $val =~ $numexp;
+               ) unless defined $val and is_number($val);
                $str =~ s/\?$i\?/$val/;
 	   }
            $str =~ s/\s//g;
@@ -1461,7 +1528,9 @@ sub get_row_value {
         /SUBSTRING/                   &&do {
                 my $start  = $structure->{"start"}->{"value"} || 1;
                 my $offset = $structure->{"length"}->{"value"} || length $value;
-                return substr($value,$start-1,$offset);
+                $value ||= '';
+                return substr($value,$start-1,$offset)
+                   if length $value >= $start-2+$offset;
         };
     }
 }
@@ -1481,8 +1550,21 @@ sub columns {
     }
 
 }
-
 sub colname2table {
+    my $self = shift;
+    my $col_name = shift;
+    return undef unless defined $col_name;
+    my $found_table;
+    for my $full_col(@{$self->{all_cols}}) {
+        my($table,$col) = $full_col =~ /^(.+)\.(.+)$/;
+        next unless $col eq $col_name;
+        $found_table = $table;
+        last;
+    }
+    return $found_table;
+}
+
+sub colname2tableOLD {
     my $self = shift;
     my $col_name = shift;
     return undef unless defined $col_name;
@@ -1490,8 +1572,10 @@ sub colname2table {
     my $table;
     my $name;
     my @cur_cols;
+print "<$col_name>";
     for my $c(@{$self->{"columns"}}) {
          $name  = $c->{"name"};
+print "[$name]\n";
          $table = $c->{"table"};
          push @cur_cols,$name;
          next unless $name eq $col_name;
@@ -1631,8 +1715,7 @@ sub do_err {
 
     $self->{"errstr"} = $err;
     warn $err if $self->{"PrintError"};
-    # print $err if $self->{"PrintError"};
-    die "\n" if $self->{"RaiseError"};
+    die "$err" if $self->{"RaiseError"};
     return undef;
 }
 
@@ -1641,6 +1724,72 @@ sub errstr {
     $self->{"errstr"};
 }
 
+sub where {
+    my $self = shift;
+    return undef unless $self->{"where_clause"};
+   $warg_num = 0;
+    return SQL::Statement::Op->new(
+        $self->{"where_clause"},
+        $self->{"tables"},
+    );
+}
+
+sub where_hash {
+    my $self = shift;
+    return $self->{where_clause};
+}
+
+
+package SQL::Statement::Op;
+
+sub new {
+    my($class,$wclause,$tables) = @_;
+    if (ref $tables->[0]) {
+        @$tables = map {$_->name} @$tables;
+    }
+    my $self = {};
+    $self->{"op"}   = $wclause->{"op"};
+    $self->{"neg"}  = $wclause->{"neg"};
+    $self->{"arg1"} = get_pred( $wclause->{"arg1"}, $tables );
+    $self->{"arg2"} = get_pred( $wclause->{"arg2"}, $tables );
+    return bless $self, $class;
+}
+sub get_pred {
+    my $arg    = shift;
+    my $tables = shift;
+    if (defined $arg->{type}) {
+        if ( $arg->{type} eq 'column') {
+            return SQL::Statement::Column->new( $arg->{value}, $tables );
+        }
+        elsif ( $arg->{type} eq 'placeholder') {
+            return SQL::Statement::Param->new( $main::warg_num++ );
+        }
+        else {
+            return $arg->{value};
+        }
+    }
+    else {
+        return SQL::Statement::Op->new( $arg, $tables );
+    }
+}
+sub op {
+    my $self = shift;
+    return $self->{"op"};
+}
+sub arg1 {
+    my $self = shift;
+    return $self->{"arg1"};
+}
+sub arg2 {
+    my $self = shift;
+    return $self->{"arg2"};
+}
+sub neg {
+    my $self = shift;
+    return $self->{"neg"};
+}
+
+
 package SQL::Statement::TempTable;
 
 sub new {
@@ -1668,10 +1817,17 @@ sub col_nums { shift->{"col_nums"} }
 sub col_names { shift->{"col_names"} }
 sub column_num  { 
     my($s,$col) = @_;
-    $col = $s->{"col_nums"}->{"$col"};
+    my $new_col = $s->{"col_nums"}->{"$col"};
+    if (! defined $new_col) {
+        my @tmp = split '~',$col;
+        $new_col = lc($tmp[0]) . '~' . uc($tmp[1]);
+        $new_col = $s->{"col_nums"}->{"$new_col"};
+    }
+    return $new_col
 }
 sub fetch_row { my $s=shift; return shift @{ $s->{"table"} } }
 
+
 package SQL::Statement::Order;
 
 sub new ($$) {
@@ -2518,7 +2674,7 @@ All rights reserved.
 
 You may distribute this module under the terms of either the GNU
 General Public License or the Artistic License, as specified in
-the Perl README file. 
+the Perl README file.
 
 
 =head1 SEE ALSO
